@@ -2,7 +2,7 @@
  * weechat-js.cpp - javascript plugin for WeeChat
  *
  * Copyright (C) 2013 Koka El Kiwi <kokakiwi@kokakiwi.net>
- * Copyright (C) 2015 Sébastien Helleu <flashcode@flashtux.org>
+ * Copyright (C) 2015-2021 Sébastien Helleu <flashcode@flashtux.org>
  *
  * This file is part of WeeChat, the extensible chat client.
  *
@@ -17,7 +17,7 @@
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
- * along with WeeChat.  If not, see <http://www.gnu.org/licenses/>.
+ * along with WeeChat.  If not, see <https://www.gnu.org/licenses/>.
  */
 
 #include <cstdlib>
@@ -39,11 +39,24 @@ WEECHAT_PLUGIN_DESCRIPTION("Support of javascript scripts");
 WEECHAT_PLUGIN_AUTHOR("Koka El Kiwi <kokakiwi@kokakiwi.net>");
 WEECHAT_PLUGIN_VERSION(WEECHAT_VERSION);
 WEECHAT_PLUGIN_LICENSE(WEECHAT_LICENSE);
-WEECHAT_PLUGIN_PRIORITY(3000);
+WEECHAT_PLUGIN_PRIORITY(4000);
 
-struct t_weechat_plugin *weechat_js_plugin;
+struct t_weechat_plugin *weechat_js_plugin = NULL;
+
+struct t_plugin_script_data js_data;
+
+struct t_config_file *js_config_file = NULL;
+struct t_config_option *js_config_look_check_license = NULL;
+struct t_config_option *js_config_look_eval_keep_context = NULL;
 
 int js_quiet = 0;
+
+struct t_plugin_script *js_script_eval = NULL;
+int js_eval_mode = 0;
+int js_eval_send_input = 0;
+int js_eval_exec_commands = 0;
+struct t_gui_buffer *js_eval_buffer = NULL;
+
 struct t_plugin_script *js_scripts = NULL;
 struct t_plugin_script *last_js_script = NULL;
 struct t_plugin_script *js_current_script = NULL;
@@ -215,7 +228,21 @@ weechat_js_exec (struct t_plugin_script *script,
         if ((ret_type == WEECHAT_SCRIPT_EXEC_STRING) && (ret_js->IsString()))
         {
             v8::String::Utf8Value temp_str(ret_js);
-            ret_value = *temp_str;
+            ret_value = (*temp_str) ? strdup(*temp_str) : NULL;
+        }
+        else if ((ret_type == WEECHAT_SCRIPT_EXEC_POINTER) && (ret_js->IsString()))
+        {
+            v8::String::Utf8Value temp_str(ret_js);
+            if (*temp_str)
+            {
+                ret_value = plugin_script_str2ptr (weechat_js_plugin,
+                                                   script->name, function,
+                                                   *temp_str);
+            }
+            else
+            {
+                ret_value = NULL;
+            }
         }
         else if ((ret_type == WEECHAT_SCRIPT_EXEC_INT) && (ret_js->IsInt32()))
         {
@@ -235,15 +262,18 @@ weechat_js_exec (struct t_plugin_script *script,
         }
         else
         {
-            weechat_printf (NULL,
-                            weechat_gettext ("%s%s: function \"%s\" must "
-                                             "return a valid value"),
-                            weechat_prefix ("error"), JS_PLUGIN_NAME,
-                            function);
+            if (ret_type != WEECHAT_SCRIPT_EXEC_IGNORE)
+            {
+                weechat_printf (NULL,
+                                weechat_gettext ("%s%s: function \"%s\" must "
+                                                 "return a valid value"),
+                                weechat_prefix ("error"), JS_PLUGIN_NAME,
+                                function);
+            }
         }
     }
 
-    if (!ret_value)
+    if ((ret_type != WEECHAT_SCRIPT_EXEC_IGNORE) && !ret_value)
     {
         weechat_printf (NULL,
                         weechat_gettext ("%s%s: error in function \"%s\""),
@@ -259,23 +289,28 @@ end:
 /*
  * Loads a javascript script.
  *
- * Returns:
- *   1: OK
- *   0: error
+ * If code is NULL, the content of filename is read and executed.
+ * If code is not NULL, it is executed (the file is not read).
+ *
+ * Returns pointer to new registered script, NULL if error.
  */
 
-int
-weechat_js_load (const char *filename)
+struct t_plugin_script *
+weechat_js_load (const char *filename, const char *code)
 {
     char *source;
 
+    /* make C compiler happy */
+    /* TODO: implement load of code in Javascript */
+    (void) code;
+
     source = weechat_file_get_content (filename);
     if (!source)
     {
         weechat_printf (NULL,
                         weechat_gettext ("%s%s: script \"%s\" not found"),
                         weechat_prefix ("error"), JS_PLUGIN_NAME, filename);
-        return 0;
+        return NULL;
     }
 
     if ((weechat_js_plugin->debug >= 2) || !js_quiet)
@@ -297,7 +332,7 @@ weechat_js_load (const char *filename)
                                          "sub-interpreter"),
                         weechat_prefix ("error"), JS_PLUGIN_NAME);
         free (source);
-        return 0;
+        return NULL;
     }
 
     /* load libs */
@@ -322,7 +357,7 @@ weechat_js_load (const char *filename)
             js_current_script = NULL;
         }
 
-        return 0;
+        return NULL;
     }
 
     free (source);
@@ -343,7 +378,7 @@ weechat_js_load (const char *filename)
                                   js_current_script);
             js_current_script = NULL;
         }
-        return 0;
+        return NULL;
     }
 
     if (!js_registered_script)
@@ -353,7 +388,7 @@ weechat_js_load (const char *filename)
                                          "found (or failed) in file \"%s\""),
                         weechat_prefix ("error"), JS_PLUGIN_NAME, filename);
         delete js_current_interpreter;
-        return 0;
+        return NULL;
     }
 
     js_current_script = js_registered_script;
@@ -372,7 +407,7 @@ weechat_js_load (const char *filename)
                               WEECHAT_HOOK_SIGNAL_STRING,
                               js_current_script->filename);
 
-    return 1;
+    return js_current_script;
 }
 
 /*
@@ -385,7 +420,7 @@ weechat_js_load_cb (void *data, const char *filename)
     /* make C++ compiler happy */
     (void) data;
 
-    weechat_js_load (filename);
+    weechat_js_load (filename, NULL);
 }
 
 /*
@@ -499,7 +534,7 @@ weechat_js_reload_name (const char *name)
                                 weechat_gettext ("%s: script \"%s\" unloaded"),
                                 JS_PLUGIN_NAME, name);
             }
-            weechat_js_load (filename);
+            weechat_js_load (filename, NULL);
             free (filename);
         }
     }
@@ -511,19 +546,42 @@ weechat_js_reload_name (const char *name)
     }
 }
 
+/*
+ * Evaluates javascript source code.
+ *
+ * Returns:
+ *   1: OK
+ *   0: error
+ */
+
+int
+weechat_js_eval (struct t_gui_buffer *buffer, int send_to_buffer_as_input,
+                 int exec_commands, const char *code)
+{
+    /* TODO: implement javascript eval */
+    (void) buffer;
+    (void) send_to_buffer_as_input;
+    (void) exec_commands;
+    (void) code;
+
+    return 1;
+}
+
 /*
  * Callback for command "/javascript".
  */
 
 int
-weechat_js_command_cb (void *data, struct t_gui_buffer *buffer,
+weechat_js_command_cb (const void *pointer, void *data,
+                       struct t_gui_buffer *buffer,
                        int argc, char **argv, char **argv_eol)
 {
-    char *ptr_name, *path_script;
+    char *ptr_name, *ptr_code, *path_script;
+    int i, send_to_buffer_as_input, exec_commands;
 
     /* make C++ compiler happy */
+    (void) pointer;
     (void) data;
-    (void) buffer;
 
     if (argc == 1)
     {
@@ -551,10 +609,16 @@ weechat_js_command_cb (void *data, struct t_gui_buffer *buffer,
             weechat_js_unload_all ();
             plugin_script_auto_load (weechat_js_plugin, &weechat_js_load_cb);
         }
-        else if (weechat_strcasecmp(argv[1], "unload"))
+        else if (weechat_strcasecmp(argv[1], "unload") == 0)
         {
             weechat_js_unload_all ();
         }
+        else if (weechat_strcasecmp (argv[1], "version") == 0)
+        {
+            plugin_script_display_interpreter (weechat_js_plugin, 0);
+        }
+        else
+            WEECHAT_COMMAND_ERROR;
     }
     else
     {
@@ -587,7 +651,8 @@ weechat_js_command_cb (void *data, struct t_gui_buffer *buffer,
                 /* load javascript script */
                 path_script = plugin_script_search_path (weechat_js_plugin,
                                                          ptr_name);
-                weechat_js_load ((path_script) ? path_script : ptr_name);
+                weechat_js_load ((path_script) ? path_script : ptr_name,
+                                 NULL);
                 if (path_script)
                     free (path_script);
             }
@@ -603,6 +668,44 @@ weechat_js_command_cb (void *data, struct t_gui_buffer *buffer,
             }
             js_quiet = 0;
         }
+        else if (weechat_strcasecmp (argv[1], "eval") == 0)
+        {
+            send_to_buffer_as_input = 0;
+            exec_commands = 0;
+            ptr_code = argv_eol[2];
+            for (i = 2; i < argc; i++)
+            {
+                if (argv[i][0] == '-')
+                {
+                    if (strcmp (argv[i], "-o") == 0)
+                    {
+                        if (i + 1 >= argc)
+                            WEECHAT_COMMAND_ERROR;
+                        send_to_buffer_as_input = 1;
+                        exec_commands = 0;
+                        ptr_code = argv_eol[i + 1];
+                    }
+                    else if (strcmp (argv[i], "-oc") == 0)
+                    {
+                        if (i + 1 >= argc)
+                            WEECHAT_COMMAND_ERROR;
+                        send_to_buffer_as_input = 1;
+                        exec_commands = 1;
+                        ptr_code = argv_eol[i + 1];
+                    }
+                }
+                else
+                    break;
+            }
+            if (!weechat_js_eval (buffer, send_to_buffer_as_input,
+                                  exec_commands, ptr_code))
+                WEECHAT_COMMAND_ERROR;
+            /* TODO: implement /javascript eval */
+            weechat_printf (NULL,
+                            _("%sCommand \"/%s eval\" is not yet implemented"),
+                            weechat_prefix ("error"),
+                            weechat_js_plugin->name);
+        }
         else
             WEECHAT_COMMAND_ERROR;
     }
@@ -615,11 +718,13 @@ weechat_js_command_cb (void *data, struct t_gui_buffer *buffer,
  */
 
 int
-weechat_js_completion_cb (void *data, const char *completion_item,
+weechat_js_completion_cb (const void *pointer, void *data,
+                          const char *completion_item,
                           struct t_gui_buffer *buffer,
                           struct t_gui_completion *completion)
 {
     /* make C++ compiler happy */
+    (void) pointer;
     (void) data;
     (void) completion_item;
     (void) buffer;
@@ -634,9 +739,11 @@ weechat_js_completion_cb (void *data, const char *completion_item,
  */
 
 struct t_hdata *
-weechat_js_hdata_cb (void *data, const char *hdata_name)
+weechat_js_hdata_cb (const void *pointer, void *data,
+                     const char *hdata_name)
 {
     /* make C++ compiler happy */
+    (void) pointer;
     (void) data;
 
     return plugin_script_hdata_script (weechat_plugin,
@@ -644,15 +751,37 @@ weechat_js_hdata_cb (void *data, const char *hdata_name)
                                        hdata_name);
 }
 
+/*
+ * Returns javascript info "javascript_eval".
+ */
+
+char *
+weechat_js_info_eval_cb (const void *pointer, void *data,
+                         const char *info_name,
+                         const char *arguments)
+{
+    const char *not_implemented = "not yet implemented";
+
+    /* make C compiler happy */
+    (void) pointer;
+    (void) data;
+    (void) info_name;
+    (void) arguments;
+
+    return strdup (not_implemented);
+}
+
 /*
  * Returns infolist with javascript scripts.
  */
 
 struct t_infolist *
-weechat_js_infolist_cb (void *data, const char *infolist_name,
-                        void *pointer, const char *arguments)
+weechat_js_infolist_cb (const void *pointer, void *data,
+                        const char *infolist_name,
+                        void *obj_pointer, const char *arguments)
 {
     /* make C++ compiler happy */
+    (void) pointer;
     (void) data;
 
     if (!infolist_name || !infolist_name[0])
@@ -661,7 +790,7 @@ weechat_js_infolist_cb (void *data, const char *infolist_name,
     if (weechat_strcasecmp (infolist_name, "javascript_script") == 0)
     {
         return plugin_script_infolist_list_scripts (weechat_js_plugin,
-                                                    js_scripts, pointer,
+                                                    js_scripts, obj_pointer,
                                                     arguments);
     }
 
@@ -673,10 +802,12 @@ weechat_js_infolist_cb (void *data, const char *infolist_name,
  */
 
 int
-weechat_js_signal_debug_dump_cb (void *data, const char *signal,
+weechat_js_signal_debug_dump_cb (const void *pointer, void *data,
+                                 const char *signal,
                                  const char *type_data, void *signal_data)
 {
     /* make C++ compiler happy */
+    (void) pointer;
     (void) data;
     (void) signal;
     (void) type_data;
@@ -690,61 +821,21 @@ weechat_js_signal_debug_dump_cb (void *data, const char *signal,
     return WEECHAT_RC_OK;
 }
 
-/*
- * Display infos about external libraries used.
- */
-
-int
-weechat_js_signal_debug_libs_cb (void *data, const char *signal,
-                                 const char *type_data, void *signal_data)
-{
-    /* make C++ compiler happy */
-    (void) data;
-    (void) signal;
-    (void) type_data;
-    (void) signal_data;
-
-    weechat_printf (NULL, "  %s (v8): %s",
-                    JS_PLUGIN_NAME, v8::V8::GetVersion());
-
-    return WEECHAT_RC_OK;
-}
-
-/*
- * Callback called when a buffer is closed.
- */
-
-int
-weechat_js_signal_buffer_closed_cb (void *data, const char *signal,
-                                    const char *type_data, void *signal_data)
-{
-    /* make C++ compiler happy */
-    (void) data;
-    (void) signal;
-    (void) type_data;
-
-    if (signal_data)
-    {
-        plugin_script_remove_buffer_callbacks (js_scripts,
-                                               (struct t_gui_buffer *)signal_data);
-    }
-
-    return WEECHAT_RC_OK;
-}
-
 /*
  * Timer for executing actions.
  */
 
 int
-weechat_js_timer_action_cb (void *data, int remaining_calls)
+weechat_js_timer_action_cb (const void *pointer, void *data,
+                            int remaining_calls)
 {
     /* make C++ compiler happy */
+    (void) data;
     (void) remaining_calls;
 
-    if (data)
+    if (pointer)
     {
-        if (data == &js_action_install_list)
+        if (pointer == &js_action_install_list)
         {
             plugin_script_action_install (weechat_js_plugin,
                                           js_scripts,
@@ -753,7 +844,7 @@ weechat_js_timer_action_cb (void *data, int remaining_calls)
                                           &js_quiet,
                                           &js_action_install_list);
         }
-        else if (data == &js_action_remove_list)
+        else if (pointer == &js_action_remove_list)
         {
             plugin_script_action_remove (weechat_js_plugin,
                                          js_scripts,
@@ -761,7 +852,7 @@ weechat_js_timer_action_cb (void *data, int remaining_calls)
                                          &js_quiet,
                                          &js_action_remove_list);
         }
-        else if (data == &js_action_autoload_list)
+        else if (pointer == &js_action_autoload_list)
         {
             plugin_script_action_autoload (weechat_js_plugin,
                                            &js_quiet,
@@ -778,11 +869,13 @@ weechat_js_timer_action_cb (void *data, int remaining_calls)
  */
 
 int
-weechat_js_signal_script_action_cb (void *data, const char *signal,
+weechat_js_signal_script_action_cb (const void *pointer, void *data,
+                                    const char *signal,
                                     const char *type_data,
                                     void *signal_data)
 {
     /* make C++ compiler happy */
+    (void) pointer;
     (void) data;
 
     if (strcmp (type_data, WEECHAT_HOOK_SIGNAL_STRING) == 0)
@@ -793,7 +886,7 @@ weechat_js_signal_script_action_cb (void *data, const char *signal,
                                       (const char *)signal_data);
             weechat_hook_timer (1, 0, 1,
                                 &weechat_js_timer_action_cb,
-                                &js_action_install_list);
+                                &js_action_install_list, NULL);
         }
         else if (strcmp (signal, "javascript_script_remove") == 0)
         {
@@ -801,7 +894,7 @@ weechat_js_signal_script_action_cb (void *data, const char *signal,
                                       (const char *)signal_data);
             weechat_hook_timer (1, 0, 1,
                                 &weechat_js_timer_action_cb,
-                                &js_action_remove_list);
+                                &js_action_remove_list, NULL);
         }
         else if (strcmp (signal, "javascript_script_autoload") == 0)
         {
@@ -809,7 +902,7 @@ weechat_js_signal_script_action_cb (void *data, const char *signal,
                                       (const char *)signal_data);
             weechat_hook_timer (1, 0, 1,
                                 &weechat_js_timer_action_cb,
-                                &js_action_autoload_list);
+                                &js_action_autoload_list, NULL);
         }
     }
 
@@ -823,22 +916,35 @@ weechat_js_signal_script_action_cb (void *data, const char *signal,
 EXPORT int
 weechat_plugin_init (struct t_weechat_plugin *plugin, int argc, char *argv[])
 {
-    struct t_plugin_script_init init;
+    char str_interpreter[64];
 
     weechat_js_plugin = plugin;
 
-    init.callback_command = &weechat_js_command_cb;
-    init.callback_completion = &weechat_js_completion_cb;
-    init.callback_hdata = &weechat_js_hdata_cb;
-    init.callback_infolist = &weechat_js_infolist_cb;
-    init.callback_signal_debug_dump = &weechat_js_signal_debug_dump_cb;
-    init.callback_signal_debug_libs = &weechat_js_signal_debug_libs_cb;
-    init.callback_signal_buffer_closed = &weechat_js_signal_buffer_closed_cb;
-    init.callback_signal_script_action = &weechat_js_signal_script_action_cb;
-    init.callback_load_file = &weechat_js_load_cb;
+    /* set interpreter name and version */
+    snprintf (str_interpreter, sizeof (str_interpreter),
+              "%s (v8)", plugin->name);
+    weechat_hashtable_set (plugin->variables, "interpreter_name",
+                           str_interpreter);
+    weechat_hashtable_set (plugin->variables, "interpreter_version",
+                           v8::V8::GetVersion());
+
+    js_data.config_file = &js_config_file;
+    js_data.config_look_check_license = &js_config_look_check_license;
+    js_data.config_look_eval_keep_context = &js_config_look_eval_keep_context;
+    js_data.scripts = &js_scripts;
+    js_data.last_script = &last_js_script;
+    js_data.callback_command = &weechat_js_command_cb;
+    js_data.callback_completion = &weechat_js_completion_cb;
+    js_data.callback_hdata = &weechat_js_hdata_cb;
+    js_data.callback_info_eval = &weechat_js_info_eval_cb;
+    js_data.callback_infolist = &weechat_js_infolist_cb;
+    js_data.callback_signal_debug_dump = &weechat_js_signal_debug_dump_cb;
+    js_data.callback_signal_script_action = &weechat_js_signal_script_action_cb;
+    js_data.callback_load_file = &weechat_js_load_cb;
+    js_data.unload_all = &weechat_js_unload_all;
 
     js_quiet = 1;
-    plugin_script_init (plugin, argc, argv, &init);
+    plugin_script_init (plugin, argc, argv, &js_data);
     js_quiet = 0;
 
     plugin_script_display_short_list (weechat_js_plugin, js_scripts);
@@ -854,7 +960,12 @@ EXPORT int
 weechat_plugin_end (struct t_weechat_plugin *plugin)
 {
     js_quiet = 1;
-    plugin_script_end (plugin, &js_scripts, &weechat_js_unload_all);
+    if (js_script_eval)
+    {
+        weechat_js_unload (js_script_eval);
+        js_script_eval = NULL;
+    }
+    plugin_script_end (plugin, &js_data);
     js_quiet = 0;
 
     /* free some data */
@@ -864,6 +975,7 @@ weechat_plugin_end (struct t_weechat_plugin *plugin)
         free (js_action_remove_list);
     if (js_action_autoload_list)
         free (js_action_autoload_list);
+    /* weechat_string_dyn_free (js_buffer_output, 1); */
 
     return WEECHAT_RC_OK;
 }
